"""victoria.storage.provider

Abstract base class of a storage provider.

Author:
    Sam Gibson 
"""
from abc import ABC, abstractmethod
from io import IOBase
from typing import Generator, Union


class StorageProvider(ABC):
    """Abstract base class of a storage provider."""
    @abstractmethod
    def store(self, data: Union[IOBase, str, bytes], key: str) -> None:
        """Store a piece of data at a given key in blob storage.

        Args:
            data: The data to store.
            key: The path within the container to store it in.
        """
        raise NotImplementedError()

    @abstractmethod
    def retrieve(self, key: str, stream: IOBase) -> None:
        """Retrieve a piece of data from a given key in blob storage.

        Args:
            key: The key within the container to get the data from.
            stream: The stream to stream the data into.
        """
        raise NotImplementedError()

    @abstractmethod
    def ls(self) -> Generator[str, None, None]:
        """List blobs within a container.

        Yields:
            str: The names of blobs within the container.
        """
        raise NotImplementedError()


"""victoria.storage.local_provider

Implementation of a StorageProvider for local file storage.

Author:
    Sam Gibson 
"""
from io import IOBase
import os
from os import path
import shutil
from typing import Generator, Union

from . import provider


class LocalStorageProvider(provider.StorageProvider):
    """Storage provider for local file storage.

    Attributes:
        container (str): The folder we're using to store files in.
    """
    def __init__(self, container: str, **kwargs):
        #pylint: disable=unused-argument
        self.container = container

    def store(self, data: Union[IOBase, str, bytes], key: str) -> None:
        self._ensure_container()
        file_path = path.join(self.container, key)
        if issubclass(type(data), IOBase):
            with open(file_path, "w") as file_handle:
                shutil.copyfileobj(data, file_handle)
        elif type(data) == str:
            with open(file_path, "w") as file_handle:
                file_handle.write(data)
        elif type(data) == bytes:
            with open(file_path, "wb") as file_handle:
                file_handle.write(data)
        else:
            raise TypeError(f"invalid data type '{type(data)}'")

    def retrieve(self, key: str, stream: IOBase) -> None:
        self._ensure_container()
        with open(path.join(self.container, key), "rb") as file_handle:
            shutil.copyfileobj(file_handle, stream)

    def ls(self) -> Generator[str, None, None]:
        self._ensure_container()
        for dirpath, _, filenames in os.walk(self.container):
            for filename in filenames:
                yield path.join(dirpath, filename)

    def _ensure_container(self) -> None:
        os.makedirs(self.container, exist_ok=True)


"""victoria.encryption.provider

Abstract base class of an encryption provider.

Author:
    Sam Gibson 
"""
from abc import ABC, abstractmethod
import os
from typing import Tuple, Optional, Union

from cryptography.hazmat.primitives.ciphers.aead import AESGCM

from .schemas import EncryptionEnvelope


class EncryptionProvider(ABC):
    """Abstract base class for an encryption provider. Provides methods
    for envelope encrypting/decrypting arbitrary byte/string sequences.
    """
    @abstractmethod
    def encrypt(self, data: bytes) -> EncryptionEnvelope:
        """Encrypt bytes to an EncryptionEnvelope.

        Args:
            data (bytes): The data to encrypt.

        Returns:
            EncryptionEnvelope: The encrypted data.
        """
        raise NotImplementedError()

    def encrypt_str(self,
                    data: str,
                    encoding: Optional[str] = "utf-8") -> EncryptionEnvelope:
        """Encrypt a string to an EncryptionEnvelope.

        Args:
            data (str): The string to encrypt.
            encoding (str, optional): The encoding of the data. Default utf-8.

        Returns:
            EncryptionEnvelope: The encrypted data.
        """
        return self.encrypt(data.encode(encoding))

    @abstractmethod
    def decrypt(self, envelope: EncryptionEnvelope) -> Union[bytes, None]:
        """Decrypt an EncryptionEnvelope to bytes.

        Please note that this can return None if the key encryption key
        is out of date.

        Args:
            envelope (EncryptionEnvelope): The envelope to decrypt.

        Returns:
            bytes: The decrypted data.
            None: If the key encryption key was out of date.
        """
        raise NotImplementedError()

    def decrypt_str(self,
                    envelope: EncryptionEnvelope,
                    encoding: Optional[str] = "utf-8") -> Union[str, None]:
        """Decrypt an EncryptionEnvelope to a string.

        Please note that this can return None if the key encryption key
        is out of date.

        Args:
            envelope (EncryptionEnvelope): The envelope to decrypt.
            encoding (str, optional): The encoding of the data. Default utf-8.

        Returns:
            str: The decrypted data.
            None: If the key encryption key was out of date.
        """
        decrypted = self.decrypt(envelope)
        if decrypted is None:
            return None
        return decrypted.decode(encoding)

    @abstractmethod
    def rotate_key(self,
                   envelope: EncryptionEnvelope,
                   version: Optional[str] = None) -> EncryptionEnvelope:
        """Rotate an EncryptionEnvelope encrypted with an old version of the
        key encryption key to a different version.

        Basically just decrypts it with the old key and reencrypts it with the 
        new key.

        Args:
            envelope (EncryptionEnvelope): The envelope to rotate.
            version (str, optional): The specific KEK version to use. Defaults
                to the latest if not given. You probably don't want to specify 
                this, as the latest will be the best version.
        
        Returns:
            EncryptionEnvelope: The newly rotated encrypted data.
        """
        raise NotImplementedError()

    def _data_encrypt(self, data: bytes) -> Tuple[bytes, bytes, bytes]:
        """Asymmetrically encrypt a piece of data, returning a tuple containing
        the encrypted data as a bytes object, the plaintext (!) encryption key 
        as a bytes object, and the nonce used to encrypt as a bytes object.

        Uses a 256-bit AES cipher in Galois Counter Mode, as recommended by 
        Google for generating data encryption keys: 
        https://cloud.google.com/kms/docs/envelope-encryption#data_encryption_keys

        Uses a 96-bit nonce length, as recommended by NIST for AES in GCM:
        https://csrc.nist.gov/publications/detail/sp/800-38d/final

        Args:
            data: The data to encrypt.

        Returns:
            Tuple:
                bytes: Encrypted data
                bytes: Plaintext (!) data encryption key -- encrypt with your 
                    key encryption key ASAP and do NOT store in plaintext!!
                bytes: 96-bit nonce value used with data encryption key when 
                    encrypting data. Under NO circumstances to be reused for
                    encryption with the same key.
        """
        key = AESGCM.generate_key(bit_length=256)
        aesgcm = AESGCM(key)
        nonce = os.urandom(12)  # 12 bytes == 96 bits
        encrypted = aesgcm.encrypt(nonce, data, associated_data=None)
        return encrypted, key, nonce

    def _data_decrypt(self, data: bytes, key: bytes, nonce: bytes) -> bytes:
        """Decrypt a piece of data with its plaintext data encryption key, and
        its nonce.

        For more information on encryption, please see the '_data_encrypt()'
        function.

        Args:
            data: The data to decrypt.
            key: The plaintext (!) data encryption key.
            nonce: The nonce used to encrypt the data with the key.

        Returns:
            bytes: The decrypted data.
        """
        aesgcm = AESGCM(key)
        return aesgcm.decrypt(nonce, data, associated_data=None)


"""victoria_encrypt

A Victoria plugin to make envelope encryption easier.

Author:
    Sam Gibson 
"""
import logging

import click
import dpath.util
from marshmallow import ValidationError
import yaml

from victoria.config import Config
from victoria.plugin import Plugin
from victoria.encryption.schemas import EncryptionEnvelopeSchema, EncryptionEnvelope


@click.group()
@click.pass_obj
def encrypt(cfg: Config):
    """Envelope encrypt data easily."""
    pass


@encrypt.command()
@click.pass_obj
@click.argument("plaintext")
def data(cfg: Config, plaintext: str):
    """Envelope encrypt a piece of data."""
    # get the provider, and encrypt the data
    provider = cfg.get_encryption()
    envelope = provider.encrypt_str(plaintext)

    # print out the encrypted data in YAML format
    print(f"data: {envelope.data}")
    print(f"iv: {envelope.iv}")
    print(f"key: {envelope.key}")
    print(f"version: {envelope.version}")


@encrypt.command()
@click.pass_obj
@click.argument("file")
@click.argument("path")
def decrypt(cfg: Config, file: str, path: str):
    """Decrypt a piece of data."""
    # get the provider
    provider = cfg.get_encryption()
    envelope = None
    try:
        loaded_dict = None

        # load the YAML file
        with open(file, 'r') as file_handle:
            loaded_dict = yaml.safe_load(file_handle)

        # scope to the envelope directly with dpath
        raw_envelope = dpath.util.get(loaded_dict, path)

        # load it with the schema
        envelope = EncryptionEnvelopeSchema().load(raw_envelope)
    except (OSError, yaml.YAMLError, ValueError, KeyError,
            ValidationError) as err:
        logging.error(err)
        raise SystemExit(1)

    decrypted = provider.decrypt_str(envelope)
    if decrypted is None:
        raise SystemExit(1)

    print(decrypted)



@encrypt.command()
@click.pass_obj
@click.argument("file")
@click.argument("path")
def rotate(cfg: Config, file: str, path: str):
    """Rotate an encryption key."""
    # get the provider
    provider = cfg.get_encryption()

    envelope = None
    try:
        loaded_dict = None

        # load the YAML file
        with open(file, 'r') as file_handle:
            loaded_dict = yaml.safe_load(file_handle)

        # scope to the envelope directly with dpath
        raw_envelope = dpath.util.get(loaded_dict, path)

        # load it with the schema
        envelope = EncryptionEnvelopeSchema().load(raw_envelope)
    except (OSError, yaml.YAMLError, ValueError, KeyError,
            ValidationError) as err:
        logging.error(err)
        raise SystemExit(1)

    # now re-encrypt it with the new key
    rotated_envelope = provider.rotate_key(envelope)

    # and print it
    print(f"data: {rotated_envelope.data}")
    print(f"iv: {rotated_envelope.iv}")
    print(f"key: {rotated_envelope.key}")
    print(f"version: {rotated_envelope.version}")


# this object is loaded by Victoria and used as the plugin entry point
plugin = Plugin(name="encrypt", cli=encrypt)



"""victoria.encryption

Package providing encryption functionality for Victoria.

Author:
    Sam Gibson 
"""
import logging

from . import provider, azure_provider
from .schemas import EncryptionProviderConfigSchema, EncryptionProviderConfig

EncryptionProvider = provider.EncryptionProvider

PROVIDERS_MAP = {"azure": azure_provider.AzureEncryptionProvider}
"""Map provider names to implementations. Used in make_provider() factory
function."""


def make_provider(provider_type: str, **kwargs) -> provider.EncryptionProvider:
    """Make an EncryptionProvider from a provider type. Call with the arguments
    you would normally provide to the specific implementation's constructor.

    make_provider("azure", vault_url="...", ...)

    Args:
        provider_type (str): The type of the provider. See PROVIDERS_MAP.
        **kwargs: The arguments for the provider implementation's constructor.

    Returns:
        EncryptionProvider: The provider of the given type.

    Raises:
        ValueError: If provider_type was invalid (i.e. not in PROVIDERS_MAP).
        TypeError: If the wrong arguments were given in kwargs.
    """
    try:
        return PROVIDERS_MAP[provider_type](**kwargs)
    except KeyError as err:
        raise ValueError(f"Invalid encryption provider_type '{provider_type}'") from err



"""victoria_config

A Victoria plugin to print the currently loaded config.

Author:
    Sam Gibson 
"""

import click
import yaml
from victoria.plugin import Plugin

# you need to use relative imports for your own modules/packages
from . import functionality
from . import schema

# this object is loaded by Victoria and used as the plugin entry point
plugin = Plugin(name="config",
                cli=functionality.config,
                config_schema=schema.ConfigSchema())



"""util

Contains various utility functions and classes used in Victoria.

Author:
    Sam Gibson 
"""

from os.path import splitext, basename


def basenamenoext(path: str) -> str:
    """Get the basename of a filepath without a file extension."""
    return splitext(basename(path))[0]


"""victoria.storage.azure_provider

Implementation of a StorageProvider for Azure Blob Storage.

Author:
    Sam Gibson 
"""
from io import IOBase
import logging
from typing import Generator, Union

from azure.common.client_factory import get_client_from_cli_profile
from azure.storage.blob import BlobServiceClient

from . import provider


class AzureStorageProvider(provider.StorageProvider):
    """Storage provider for Azure Blob Storage.

    Attributes:
        main_client (BlobServiceClient): The Azure blob service client.
        client (ContainerClient): The Azure API blob client.

    Args:
        container (str): The storage container to use.
        auth_via_cli (bool): Whether we should authenticate via the Azure CLI.
        account_name (str): The name of the storage account. Only used if auth via CLI.
        connection_string (str): The connection string to the Azure Storage account. Only used if not auth via CLI.
    """
    def __init__(self,
                 container: str,
                 auth_via_cli: bool = False,
                 account_name: str = "",
                 connection_string: str = ""):
        if auth_via_cli:
            if not account_name:
                logging.error(
                    "ERROR: Need to specify 'account_name' for "
                    "Azure storage provider when authenticating via Azure CLI")
                raise SystemExit(1)
            try:
                self.main_client = get_client_from_cli_profile(
                    BlobServiceClient,
                    account_url=f"https://{account_name}.blob.core.windows.net/"
                )

            except ImportError as err:
                logging.error(
                    "ERROR: Unable to authenticate via Azure CLI, have you "
                    "logged in with 'az login'?")
                raise SystemExit(1) from err
        else:
            if not connection_string:
                logging.error(
                    "ERROR: Need to specify 'connection_string' for "
                    "Azure storage provider when not authenticating via Azure CLI"
                )
                raise SystemExit(1)
            self.main_client = BlobServiceClient.from_connection_string(
                connection_string)

        self.client = self.main_client.get_container_client(container)

    def store(self, data: Union[IOBase, str, bytes], key: str) -> None:
        blob_client = self.client.get_blob_client(key)
        blob_client.upload_blob(data)

    def retrieve(self, key: str, stream: IOBase):
        blob_client = self.client.get_blob_client(key)
        download_stream = blob_client.download_blob()
        stream.write(download_stream.readall())

    def ls(self) -> Generator[str, None, None]:
        for blob in self.client.list_blobs():
            yield blob.name


#!/usr/bin/env python3
"""victoria

Victoria is the SRE toolbelt -- a single command with multiple pluggable
subcommands for automating any number of 'toil' tasks that inhibit SRE
productivity.

Author:
    Sam Gibson 
"""
import argparse

import click

from .. import config
from .. import plugin

# used for making it so we can use both -h and --help for help text.
CONTEXT_SETTINGS = {"help_option_names": ["-h", "--help"]}

HELP_TEXT = """
V.I.C.T.O.R.I.A.

\b
Very
Important
Commands for
Toil
Optimization:
Reducing
Inessential
Activities

Victoria is the SRE toolbelt -- a single command with multiple pluggable
subcommands for automating any number of 'toil' tasks that inhibit SRE
productivity.
"""

VERSION_NUMBER = "v0.8.2"
"""The version number of the application, to print when calling with --version."""


class VictoriaCLI(click.MultiCommand):
    """VictoriaCLI overrides click.MultiCommand to support loading click
    commands from python files. This is the bread and butter of the plugin
    system."""
    def __init__(self, name=None, **attrs):
        click.MultiCommand.__init__(self, name, **attrs)
        self.plugins = plugin.load_all()

    def list_commands(self, ctx):
        """List the available subcommands."""
        return [plgn.name for plgn in self.plugins]

    def get_command(self, ctx, cmd_name):
        """Get a subcommand from the list of installed plugins."""
        for plgn in self.plugins:
            # if the command name matches a loaded plugin name
            if plgn.name == cmd_name:
                # if the plugin has a config schema, load its config
                if plgn.config_schema and ctx.obj:
                    cfg = config.load_plugin_config(plgn, ctx.obj)
                    if not cfg:
                        # if there was an error loading the config, exit
                        raise SystemExit(1)
                    # HACK: patch the CLI context with the plugin config object
                    # this is a bit hacky, but it works
                    plgn.cli.context_settings = {"obj": cfg}
                return plgn.cli
        return None


@click.command(cls=VictoriaCLI,
               context_settings=CONTEXT_SETTINGS,
               help=HELP_TEXT)
@click.option("-c",
              "--config-file",
              default=config.get_config_loc(),
              metavar="FILE",
              help="The config file to load.")
@click.version_option(version=VERSION_NUMBER)
@click.pass_context
def cli(ctx, config_file):
    """This is the main CLI of the application. It uses VictoriaCLI to call
    loaded plugins based on subcommand name."""
    #pylint: disable=unused-argument
    if config_file is None:
        raise SystemExit(1)


def main():
    # HACK: use argparse to get the config-file argument, as we need it in
    # VictoriaCLI methods before it would be parsed in the click cli() command...
    # this is utter filth, but I couldn't think of a better way to do it,
    # and it works transparent to the user
    parser = argparse.ArgumentParser(description="", add_help=False)
    parser.add_argument("-c", "--config-file", default=config.get_config_loc())
    parsed_args, _ = parser.parse_known_args()

    # make sure a config already exists if a custom one was not specified
    if parsed_args.config_file == config.get_config_loc():
        config.ensure()

    # load the config
    cfg = config.load(parsed_args.config_file)

    # execute the main CLI, passing the config in through the context
    cli.main(obj=cfg)


if __name__ == "__main__":
    main()


"""schema.py

This is used to store a schema for the plugin's config section.

Author:
    Sam Gibson 
"""

from marshmallow import Schema, fields, post_load, validate


class ConfigConfig:
    """ConfigConfig is the config for the config printing plugin.

    Attributes:
        indent (int): How much to indent the YAML by. Can be between 1 and 10.
    """
    def __init__(self, indent: int):
        self.indent = indent


class ConfigSchema(Schema):
    """Marshmallow schema for the config printing plugin's config section."""
    indent = fields.Int(validate=validate.Range(min=1, max=10))

    @post_load
    def make_config(self, data, **kwargs):
        return ConfigConfig(**data)


"""victoria.encryption.azure_provider

Implementation of an EncryptionProvider for Azure Key Vault.

Author:
    Sam Gibson 
"""
import base64
import logging
import os
from typing import Optional, Union

from azure.common.client_factory import get_client_from_cli_profile
from azure.identity import ClientSecretCredential
from azure.keyvault.keys import KeyClient
from azure.keyvault.keys.crypto import CryptographyClient, EncryptionAlgorithm

from .provider import EncryptionProvider, EncryptionEnvelope

ENCRYPTION_ALGORITHM = EncryptionAlgorithm.rsa_oaep_256

TENANT_ID_ENVVAR = "AZURE_TENANT_ID"
CLIENT_ID_ENVVAR = "AZURE_CLIENT_ID"
CLIENT_SECRET_ENVVAR = "AZURE_CLIENT_SECRET"  # nosec


class AzureEncryptionProvider(EncryptionProvider):
    """An EncryptionProvider implementation for Azure Key Vault.

    To authenticate, please provide Azure AD service principal info as 
    'tenant_id', 'client_id', and 'client_secret' kwargs, or as
    AZURE_TENANT_ID, AZURE_CLIENT_ID, and AZURE_CLIENT_SECRET environment
    variables.

    Attributes:
        tenant_id (str): The tenant ID of the Azure SP to connect with.
        client_id (str): The client ID of the Azure SP to connect with.
        client_secret (str): The client secret of the Azure SP to connect with.
        key_client (KeyClient): Azure Key Vault key client.
        crypto_client (CryptographyClient): Azure Key Vault crypto client.

    Args:
        vault_url (str): The URL of the key vault to connect to.
        key (str): The name of the key encryption key to use for envelope encryption.
        auth_via_cli (bool, kwarg): If we should auth via Azure CLI.
        **kwargs: Authentication information.

    Raises:
        TypeError: If authentication information is not provided correctly.
    """
    def __init__(self, vault_url: str, key: str, **kwargs) -> None:
        auth_via_cli = bool(kwargs.pop("auth_via_cli", False))
        if auth_via_cli:
            try:
                self.key_client = get_client_from_cli_profile(
                    KeyClient, vault_url=vault_url)
                self.key_encryption_key = self.key_client.get_key(key)
                self.crypto_client = get_client_from_cli_profile(
                    CryptographyClient, key=self.key_encryption_key)
            except ImportError as err:
                logging.error(
                    "ERROR: Unable to authenticate via Azure CLI, have you "
                    "logged in with 'az login'?")
                raise SystemExit(1) from err
        else:
            tenant_id = kwargs.pop("tenant_id", os.getenv(TENANT_ID_ENVVAR))
            client_id = kwargs.pop("client_id", os.getenv(CLIENT_ID_ENVVAR))
            client_secret = kwargs.pop("client_secret",
                                       os.getenv(CLIENT_SECRET_ENVVAR))
            if tenant_id is None or client_id is None or client_secret is None:
                raise TypeError(
                    "Please specify tenant_id, client_id, and client_secret "
                    "in config or in environment variables as in "
                    "https://github.com/Azure/azure-sdk-for-python/tree/master/sdk/identity/azure-identity#service-principal-with-secret"
                )
            self.cred = ClientSecretCredential(tenant_id, client_id,
                                               client_secret)
            self.key_client = KeyClient(vault_url,
                                        credential=self.cred,
                                        logger=None)
            self.key_encryption_key = self.key_client.get_key(key)
            self.crypto_client = CryptographyClient(self.key_encryption_key,
                                                    self.cred)

    # overrides EncryptionProvider.encrypt()
    def encrypt(self, data: bytes) -> EncryptionEnvelope:
        # encrypt the data locally, generating a data key and a nonce
        ciphertext, data_key, nonce = self._data_encrypt(data)

        # encrypt the data key using the key from the vault
        result = self.crypto_client.encrypt(ENCRYPTION_ALGORITHM, data_key)
        del data_key  # we don't wanna keep this around after we've encrypted it
        encrypted_data_key = result.ciphertext

        # encode to base64 for storage/transmission
        b64_ciphertext = base64.b64encode(ciphertext).decode("utf-8")
        b64_encrypted_data_key = base64.b64encode(encrypted_data_key).decode(
            "utf-8")
        b64_nonce = base64.b64encode(nonce).decode("utf-8")

        return EncryptionEnvelope(b64_ciphertext, b64_encrypted_data_key,
                                  b64_nonce,
                                  self.key_encryption_key.properties.version)

    # overrides EncryptionProvider.decrypt()
    def decrypt(self, envelope: EncryptionEnvelope) -> Union[bytes, None]:
        if envelope.version != self.key_encryption_key.properties.version:
            logging.error(
                "Encryption key version %s is out of "
                "date, please re-encrypt with 'victoria encrypt rotate'",
                envelope.version)
            return None

        # decode from base64
        ciphertext = base64.b64decode(envelope.data)
        encrypted_data_key = base64.b64decode(envelope.key)
        nonce = base64.b64decode(envelope.iv)

        # decrypt the data key
        result = self.crypto_client.decrypt(ENCRYPTION_ALGORITHM,
                                            encrypted_data_key)
        data_key = result.plaintext

        # decrypt the data locally using the nonce and decrypted data key
        return self._data_decrypt(ciphertext, data_key, nonce)

    # overrides EncryptionProvider.rotate_key()
    def rotate_key(self,
                   envelope: EncryptionEnvelope,
                   version: Optional[str] = None) -> EncryptionEnvelope:
        old_key = self.key_client.get_key(self.key_encryption_key.name,
                                          version=envelope.version)
        old_crypto_client = CryptographyClient(old_key, self.cred)

        # decode from base64
        ciphertext = base64.b64decode(envelope.data)
        encrypted_data_key = base64.b64decode(envelope.key)
        nonce = base64.b64decode(envelope.iv)

        # decrypt the data key
        result = old_crypto_client.decrypt(ENCRYPTION_ALGORITHM,
                                           encrypted_data_key)
        data_key = result.plaintext

        # decrypt the data locally using the nonce and decrypted data key
        decrypted_data = self._data_decrypt(ciphertext, data_key, nonce)

        # now re-encrypt with the latest key encryption key
        return self.encrypt(decrypted_data)



"""victoria_store

A Victoria plugin to store things in cloud storage.

Author:
    Sam Gibson 
"""
import io
import logging
import os
import sys
from os import path

import click
import yaml

from victoria import storage
from victoria.plugin import Plugin


@click.group()
@click.pass_context
@click.argument("backend")
def store(ctx, backend: str):
    """Store files/data in cloud storage."""
    # get the victoria config, so we can get cloud storage details from it
    v_cfg = ctx.obj
    ctx.obj = v_cfg.get_storage(backend)


@store.command()
@click.argument("filename", type=click.File('rb'))
@click.argument("key", type=str, required=False)
@click.pass_obj
def put(provider: storage.StorageProvider, filename, key):
    """Put a file at FILENAME into cloud storage, under an optional KEY."""
    key_name = key if key else path.basename(filename.name)
    logging.info(f"Uploading '{filename.name}' to object '{key_name}'")
    provider.store(filename, key_name)


@store.command()
@click.argument("key", type=str, required=False)
@click.pass_obj
def get(provider: storage.StorageProvider, key):
    """Get a file from KEY in cloud storage and print it to stdout."""
    stream = io.BytesIO()
    provider.retrieve(key, stream)
    print(stream.getvalue().decode("utf-8"))


@store.command()
@click.pass_obj
def ls(provider: storage.StorageProvider):
    """List files from cloud storage."""
    for blob in provider.ls():
        print(blob)


# this object is loaded by Victoria and used as the plugin entry point
plugin = Plugin(name="store", cli=store)



"""config

Contains various classes and methods for loading the config of Victoria.

Author:
    Sam Gibson 
"""
import io
import logging.config
import os
from os import path
from typing import Mapping

import appdirs
import click
from marshmallow import Schema, fields, post_load, ValidationError, EXCLUDE
import pkg_resources
import yaml

from .plugin import Plugin
from . import storage
from . import encryption

APP_NAME = "victoria"
"""What the app is called."""

APP_AUTHOR = "GlasswallSRE"
"""Who the author of the application is."""

DEFAULT_CONFIG_NAME = "victoria.yaml"
"""The default filename of the config file. This will be loaded if no other
file is given."""

EXAMPLE_CONFIG_FILE = "victoria_example.yaml"
"""The example config that will be created if a config wasn't found."""


def get_config_loc() -> str:
    """Get the path to the config file."""
    return path.join(appdirs.user_config_dir(APP_NAME, APP_AUTHOR),
                     DEFAULT_CONFIG_NAME)


def ensure() -> None:
    """Ensure that a config exists in the location."""
    if not path.exists(get_config_loc()):
        print("Didn't find config file -- installing default to "
              f"{get_config_loc()}")

        # make all of the directories to the config file
        try:
            os.makedirs(path.dirname(get_config_loc()))
        except FileExistsError:
            pass

        # now write the example config file to the config location
        with open(get_config_loc(), "w") as cfg_file:
            # load the example config from package resources and replace any
            # funky windows line endings that may be in it
            example_cfg_file = pkg_resources.resource_string(
                "victoria",
                EXAMPLE_CONFIG_FILE).decode("utf-8").replace("\r", "")
            cfg_file.write(example_cfg_file)


class ConfigSchema(Schema):
    """Marshmallow schema for the Config object."""
    logging_config = fields.Dict(required=True)
    storage_providers = fields.Dict(required=False, default={}, missing={})
    encryption_provider = fields.Nested(
        encryption.EncryptionProviderConfigSchema,
        required=False,
        allow_none=True)
    plugins_config_location = fields.Mapping(keys=fields.Str(),
                                             values=fields.Str(),
                                             default={})
    plugins_config = fields.Dict(required=False, default={}, missing={})

    @post_load
    def make_config_obj(self, data, **kwargs):
        #pylint: disable=unused-argument
        return Config(**data)


CONFIG_SCHEMA = ConfigSchema(unknown=EXCLUDE)
"""Instance of ConfigSchema used for validating loaded configs."""


class Config:
    """Config is used for storing deserialized values from Config files.

    Attributes:
        logging_config (dict): The config to use for logging.
        plugins_config (dict): The config to use for plugins.
        plugins_config_location (Mapping[str, str]): Config file location overrides for plugins.
        storage_providers (dict): Data used for connecting to storage.
        encryption_provider (EncryptionProviderConfig): Config for the encryption provider.
    """
    def __init__(
            self,
            logging_config: dict,
            plugins_config: dict = None,
            plugins_config_location: Mapping[str, str] = None,
            storage_providers: dict = None,
            encryption_provider: encryption.EncryptionProviderConfig = None):
        self.logging_config = logging_config
        logging.config.dictConfig(logging_config)
        self.plugins_config = plugins_config
        self.plugins_config_location = {} if plugins_config_location is None else plugins_config_location
        self.storage_providers = storage_providers
        self.encryption_provider = encryption_provider

    def __eq__(self, other):
        if isinstance(self, other.__class__):
            return self.logging_config == other.logging_config \
                and self.plugins_config == other.plugins_config \
                and self.plugins_config_location == other.plugins_config_location \
                and self.storage_providers == other.storage_providers \
                and self.encryption_provider == other.encryption_provider
        return False

    def as_dict(self) -> dict:
        d = vars(self)
        for k, v in d.items():
            if hasattr(v, "__dict__"):
                d[k] = vars(v)
        return d

    def get_storage(self, provider: str) -> storage.StorageProvider:
        """Get the storage provider of a given type.

        Args:
            provider (str): The provider type to get. Any key of 'storage_providers'
                mapping in Victoria core config.

        Returns:
            StorageProvider: The storage provider.

        Raises:
            ValueError: if the provider type was invalid.
        """
        try:
            return storage.make_provider(provider,
                                         **self.storage_providers[provider])
        except KeyError as err:
            logging.error(f"unknown storage provider type '{provider}'")
            raise ValueError(f"unknown storage provider type '{provider}'") from err
        except (ValueError, TypeError) as err:
            logging.error(err)
            raise

    def get_encryption(self) -> encryption.EncryptionProvider:
        """Get the encryption provider.

        Returns:
            EncryptionProvider: The encryption provider.
        """
        try:
            return encryption.make_provider(self.encryption_provider.provider,
                                            **self.encryption_provider.config)
        except (ValueError, TypeError) as err:
            logging.error(err)
            raise


pass_config = click.make_pass_decorator(Config)
"""Decorator for passing the Victoria config to a command."""


def _print_validation_err(err: ValidationError, name: str) -> None:
    """Internal function used for printing a validation error in the Schema.

    Args:
        err (ValidationError): The error to log.
        name (str): A human-readable identifier for the Schema data source. 
            Like a filename.
    """
    # build up a string for each error
    log_str = []
    log_str.append(f"Error validating config '{name}':")
    for field_name, err_msgs in err.messages.items():
        log_str.append(f"{field_name}: {err_msgs}")

    # print the joined up string
    print(" ".join(log_str))


def load_plugin_config(plugin: Plugin, cfg: Config) -> object:
    """Load the config of a plugin from the main Victoria config.

    Args:
        plugin (Plugin): The plugin to load the config for.
        cfg (Config): The config file to load the config from.

    Returns:
        object: The loaded config object. It will be the same type as whatever
            the plugin's config marshmallow schema will marshal it to, or None
            if there was some error loading the plugin config.

    Raises:
        ValueError: if the plugin didn't have a config schema.
    """
    # we don't want to try to load a config from the config if it wasn't loaded
    if cfg is None:
        logging.warning(f"Can't load '{plugin.name}' config: victoria config "
                        "not loaded.")
        return None

    if plugin.config_schema is None:
        raise ValueError(f"Can't load plugin config: plugin '{plugin.name}' "
                         "did not have config schema")

    raw_config = {}

    # check to see if there's a location override for the config
    if cfg.plugins_config is not None and plugin.name in cfg.plugins_config:
        # otherwise use the one in the main config
        if not cfg.plugins_config:
            logging.error("Can't load plugin config: config did not "
                          "have 'plugins_config' section")
            return None
        raw_config = cfg.plugins_config[plugin.name]
    elif plugin.name in cfg.plugins_config_location:
        loc = cfg.plugins_config_location[plugin.name]
        raw_config = _handle_config_file_override(loc, cfg)
    else:
        logging.error(
            "Can't load plugin config: "
            f"config did not have section for plugin '{plugin.name}' "
            "and no location was given in 'plugins_config_location'")
        return None

    # try validating the contents of the plugin section with the schema
    try:
        loaded_config = plugin.config_schema.load(raw_config)
        loaded_config = _inject_core_config(cfg, loaded_config)
        return loaded_config
    except ValidationError as err:
        # if the loaded YAML wasn't a valid config
        _print_validation_err(err, f"plugins_config.{plugin.name}")
        return None


def _handle_config_file_override(override_loc: str, cfg: Config) -> object:
    """Handle a plugin config override from the 'plugins_config_location'
    section of the Victoria core config.

    Attempts to get config from a storage provider.

    Args:
        override_loc (str): The location of the file. {provider}://{file_path}
        cfg (Config): Core Victoria config.

    Returns:
        object: The config loaded from the storage provider.
    """
    provider_type, file_path = tuple(override_loc.split("://"))
    provider = cfg.get_storage(provider_type)
    config_file = io.BytesIO()
    provider.retrieve(file_path, config_file)
    config_str = config_file.getvalue().decode("utf-8")
    return yaml.safe_load(config_str)


def _inject_core_config(core_cfg: Config, loaded_config: object) -> object:
    """Inject the core config into a loaded plugin config object."""
    if isinstance(loaded_config, dict):
        # if it's a dict, add it as a key
        loaded_config["victoria_config"] = core_cfg
    else:
        # otherwise assume it's an object
        loaded_config.victoria_config = core_cfg
    return loaded_config


def load(config_path: str) -> Config:
    """Load a config file from a given path and make sure it's valid.

    If an error occurred, it will print it and return None.

    Args:
        config_path (str): The path to the YAML config file.

    Returns:
        Config: The loaded config file, or None if an error occurred.
    """
    try:
        with open(config_path, "r") as config_file:
            raw_config = yaml.safe_load(config_file)
            loaded_config = CONFIG_SCHEMA.load(raw_config)
            return loaded_config
    except OSError as err:
        # if there was an error opening the file
        print("Error opening config file: " + str(err))
        return None
    except yaml.YAMLError as err:
        # if the loaded YAML was invalid YAML
        print("Error in config file: " + str(err))
        return None
    except ValidationError as err:
        # if the loaded YAML wasn't a valid config
        _print_validation_err(err, config_path)
        return None
    except (ValueError, TypeError, AttributeError, ImportError) as err:
        # if the logging config was invalid
        print("Unable to load logging config: " + str(err))
        return None



"""setup.py

Used for installing Victoria via pip.
"""

from setuptools import setup, find_packages


def repo_file_as_string(file_path: str) -> str:
    with open(file_path, "r") as repo_file:
        return repo_file.read()


setup(
    dependency_links=[],
    install_requires=[
        "appdirs>=1.4.3", "click>=7.0", "marshmallow>=3.2.1", "pyyaml>=5.1.2",
        "azure-storage-blob>=12.3.0", "cryptography>=2.8",
        "azure-keyvault>=4.0.0", "azure-identity>=1.3.0", "dpath>=2.0.1"
    ],
    name="victoria",
    version="0.8.2",
    description=
    "Automation toolbelt -- a single command with multiple pluggable subcommands for automating any number of 'toil' tasks that inhibit productivity.",
    long_description=repo_file_as_string("README.md"),
    long_description_content_type="text/markdown",
    author="Glasswall SRE",
    author_email="sre@glasswallsolutions.com",
    url="https://github.com/glasswall-sre/victoria",
    classifiers=[
        "Development Status :: 4 - Beta", "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8", "Topic :: Terminals",
        "Topic :: Utilities"
    ],
    keywords="cli automation sre plugin",
    packages=find_packages(".") +
    find_packages("example_plugins/victoria_config") +
    find_packages("example_plugins/victoria_store") +
    find_packages("example_plugins/victoria_encrypt"),
    package_dir={
        "victoria_config": "example_plugins/victoria_config/victoria_config",
        "victoria_store": "example_plugins/victoria_store/victoria_store",
        "victoria_encrypt": "example_plugins/victoria_encrypt/victoria_encrypt"
    },
    entry_points="""
        [console_scripts]
        victoria=victoria.script.victoria:main
    """,
    python_requires=">=3.6",
    include_package_data=True,
    package_data={"victoria": ["victoria_example.yaml"]})



"""functionality.py

This is where the main functionality of the config plugin is implemented.

Author:
    Sam Gibson 
"""

import click
import yaml
import victoria.config
from victoria.config import Config

from . import schema


@click.group()
@click.pass_obj
def config(cfg: schema.ConfigConfig):
    """Print the loaded config, and the path to the default config file."""
    pass


@config.command()
@click.pass_obj
def path(cfg: schema.ConfigConfig):
    """Print the path to the config file Victoria uses."""
    print(victoria.config.get_config_loc())


@config.command()
@click.pass_obj
def view(cfg: schema.ConfigConfig):
    """Print the current loaded config and exit."""
    core_config = cfg.victoria_config
    print(yaml.safe_dump(core_config.as_dict(), indent=cfg.indent))


"""victoria.encryption.schemas

Marshmallow schemas for the encryption functionality.

Author:
    Sam Gibson 
"""
from marshmallow import Schema, fields, post_load


class EncryptionEnvelopeSchema(Schema):
    """Marshmallow schema for an EncryptionEnvelope.

    Fields:
        data (str): The base64-encoded encrypted data.
        key (str): The base64-encoded encrypted data encryption key.
        iv (str): The base64-encoded nonce.
        version (str): The version of the key encryption key used.
    """
    data = fields.Str()
    key = fields.Str()
    iv = fields.Str()
    version = fields.Str()

    @post_load
    def make_encryption_envelope(self, data, **kwargs):
        #pylint: disable=unused-argument
        return EncryptionEnvelope(**data)


class EncryptionEnvelope:
    """Stores data that has been envelope encrypted.

    Attributes:
        data (str): The base64-encoded encrypted data.
        key (str): The base64-encoded encrypted data encryption key.
        iv (str): The base64-encoded nonce.
        version (str): The version of the key encryption key used.

    See:
        https://cloud.google.com/kms/docs/envelope-encryption
    """
    def __init__(self, data: str, key: str, iv: str, version: str) -> None:
        self.data = data
        self.key = key
        self.iv = iv
        self.version = version


class EncryptionProviderConfigSchema(Schema):
    """Marshmallow schema for the encryption provider config in Victoria core
    config.

    Fields:
        provider (str): The type of the provider.
        config (Mapping[str, str]): Params to pass to the constructor of the
            encryption provider implementation.
    """
    provider = fields.Str(required=True)
    config = fields.Mapping(keys=fields.Str(), values=fields.Raw(), missing={})

    @post_load
    def make_encryption_provider_config(self, data, **kwargs):
        #pylint: disable=unused-argument
        return EncryptionProviderConfig(**data)


class EncryptionProviderConfig:
    """Stores config for Victoria's encryption providing functionality.

    Attributes:
        provider (str): The type of the provider.
        config (Mapping[str, str]): Params to pass to the constructor of the
            encryption provider implementation.
    """
    def __init__(self, provider: str, config: dict) -> None:
        self.provider = provider
        self.config = config

    @classmethod
    def to_yaml(cls, dumper, data):
        #pylint: disable=unused-argument
        return vars(data)


"""plugin

Contains functions for loading Victoria plugins at runtime.

Author:
    Sam Gibson 
"""

import importlib
import logging
import pkgutil
from typing import List

import click
from marshmallow import Schema


class Plugin:
    """Plugin is the API for creating a Victoria plugin. It defines plugin name,
    CLI entry point, and optionally a schema for custom plugin config.

    Attributes:
        name (str): The name of the plugin. This will be the subcommand name.
        cli (click.Command): The click entrypoint of the plugin command.
        config_schema (marshmallow.Schema): The schema to use to validate config (OPTIONAL).
    """
    def __init__(self,
                 name: str,
                 cli: click.Command,
                 config_schema: Schema = None) -> None:
        self.name = name
        self.cli = cli
        self.config_schema = config_schema

    def __eq__(self, other):
        if isinstance(self, other.__class__):
            return self.name == other.name and self.cli == other.cli \
                and self.config_schema == other.config_schema
        return False


def load(plugin_name: str) -> Plugin:
    """Load a plugin by name.

    Plugins are installed python modules with a name starting with 'victoria_'.
    For example: 'victoria_config'. Plugin packages should define a plugin
    object in their __init__.py and use it to wire up plugin functionality.

    A plugin can optionally define a marshmallow schema that can be used for
    validating its config. Plugin configs will be located in the
    'plugins_config' object of the Victoria config, under an object with the
    same name as the plugin.

    Args:
        plugin_name (str): The name of the plugin package to load. Plugin packages
            must be in the format "victoria_{plugin_name}".

    Returns:
        Plugin: the plugin object as loaded from the module, or None if there was
            some error whilst loading.
    """
    try:
        spec = importlib.util.find_spec(f"victoria_{plugin_name}")
        if spec is None:
            logging.error(
                "Error loading plugin 'victoria_%s': module spec not found",
                plugin_name)
            return None
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        plugin_obj = getattr(module, "plugin")
        if type(plugin_obj) is not Plugin:
            # check that the plugin object is of the right type
            logging.error(
                "Error loading plugin 'victoria_%s': 'plugin' "
                "must be of type 'Plugin'", plugin_name)
            return None
    except (ValueError, ModuleNotFoundError, AttributeError) as err:
        # if there was some error finding the module spec or getting
        # the plugin object from the loaded module
        logging.error("Error loading plugin 'victoria_%s': %s", plugin_name,
                      err)
        return None
    return plugin_obj


def load_all() -> List[Plugin]:
    """Load all plugins that can be found.

    Returns:
        List[Plugin]: The list of plugins successfully loaded.
    """
    plugins = []
    already_loaded_names = []
    for plgn in ls():
        loaded_plugin = load(plgn)

        if loaded_plugin is not None:
            if loaded_plugin.name in already_loaded_names:
                logging.error(
                    "Error loading plugin 'victoria_%s"
                    ": a plugin with the same name is already loaded",
                    loaded_plugin.name)
                continue

            already_loaded_names.append(loaded_plugin.name)
            plugins.append(loaded_plugin)
    return plugins


def ls() -> List[str]:
    """Get a list of plugins installed on the system.

    Returns:
        List[str]: The list of plugin package names installed. Each plugin name
            will be in the list without the "victoria_" prefix.
    """
    return [
        name[len("victoria_"):] for _, name, ispkg in pkgutil.iter_modules()
        if name.startswith("victoria_") and ispkg
    ]